# Copyright 2021 Canonical Limited.
#
# This file is part of juju-verify.
#
# juju-verify is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# juju-verify is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see https://www.gnu.org/licenses/.
"""Juju-verify verification result."""
import logging
import os
from enum import Enum
from functools import total_ordering
from typing import List

logger = logging.getLogger(__name__)


@total_ordering
class Severity(Enum):
    """Collection of possible Result's severities."""

    OK = 10
    WARN = 20
    UNSUPPORTED = 30
    FAIL = 40

    def __lt__(self, other: object) -> bool:
        """Perform "less than" comparison with other Severity instances."""
        if not isinstance(other, Severity):
            return NotImplemented

        return self.value < other.value  # pylint: disable=W0143


class Partial:
    """Class representing partial result.

    Generally, instances of this class are used in Result class to represent (partial)
    results of individual checks ran during the whole verification process.
    """

    def __init__(self, severity: Severity, message: str):
        """Initialize Partial instance.

        :param severity: Severity of the partial result.
        :param message: Additional information about the result
        """
        self.severity = severity
        self.message = message

    def __str__(self) -> str:
        """Return string representation of the Partial class instance."""
        return f'[{self.severity.name}] {self.message}'

    def __eq__(self, other: object) -> bool:
        """Perform equal comparison with another Partial instance."""
        if not isinstance(other, Partial):
            return NotImplemented
        return self.severity == other.severity and self.message == other.message


class Result:
    """Convenience class that represents result of the check.

    Each juju-verify check should return Result instance with at least one partial
    result in self.partials. Multiple partial result are good idea if, for example, the
    check runs against multiple units.
    """

    VERBOSE_MAP = {
        Severity.OK: 'OK (All checks passed)',
        Severity.WARN: 'OK (Checks passed with warnings)',
        Severity.UNSUPPORTED: 'Failed (Targeted charms are not supported)',
        Severity.FAIL: 'Failed',
    }

    def __init__(self, severity: Severity = Severity.OK, message: str = ''):
        """Initialize result instance.

        Initial arguments 'severity' and 'message' will automatically create Partial
        result that will be stored in the self.partials. Additional partial results can
        be added via self.add_partial_result.
        For convenience, it's possible to initiate empty Result, however keep in mind
        that returning empty Result is not very helpful.

        :param severity: Severity of the result
        :param message: Additional information about the result.
        """
        self.partials: List[Partial] = []
        if message:
            self.partials.append(Partial(severity, message))

    def __str__(self) -> str:
        """Return formatted string representing the result.

        Note: If the Result instance has no Partial results in self.partials, returned
        string will be error message.
        """
        if not self.partials:
            return ('No result or additional information. This may be a bug in '
                    '"juju-verify".')
        output = f'Checks:{os.linesep}'
        for partial in self.partials:
            output += f'{partial}{os.linesep}'
        output += os.linesep

        max_severity = max(partial.severity for partial in self.partials)
        output += f'Overall result: {self.VERBOSE_MAP.get(max_severity)}'
        return output

    def __add__(self, other: object) -> 'Result':
        """Perform "add" operation with another Result instance."""
        if not isinstance(other, Result):
            return NotImplemented
        new_obj = Result()

        for partial in self.partials + other.partials:
            new_obj.partials.append(partial)

        return new_obj

    def __iadd__(self, other: object) -> 'Result':
        """Perform "inplace add" operation with another Result instance."""
        if not isinstance(other, Result):
            return NotImplemented

        for partial in other.partials:
            self.partials.append(partial)
        return self

    def __eq__(self, other: object) -> bool:
        """Compare two Result instances."""
        if not isinstance(other, Result):
            return NotImplemented

        return self.partials == other.partials and self.success == other.success

    @property
    def success(self) -> bool:
        """Return overall Result's success.

        The success is calculated based on the highest Severity value from all the
        partial results in self.partials. Following is the map for overall success:

        Highest severity is OK -> Overall success is True
        Highest severity is WARN -> Overall success is True
        Highest severity is UNSUPPORTED -> Overall success is False
        Highest severity is FAIL -> Overall success is False

        Note: If the Result instance has no Partial results in self.partials, overall
        success will be True.
        """
        return all(partial.severity < Severity.UNSUPPORTED for partial in self.partials)

    @property
    def empty(self) -> bool:
        """Return True if result does not contain any partial results."""
        return not bool(self.partials)

    def add_partial_result(self, severity: Severity, message: str) -> None:
        """Add partial result to this instance."""
        self.partials.append(Partial(severity, message))


def aggregate_results(*results: Result) -> Result:
    """Return aggregate value of multiple results."""
    result_list = list(results)
    final_result = result_list.pop(0)

    for result in result_list:
        final_result += result

    return final_result



# Copyright 2021 Canonical Limited.
#
# This file is part of juju-verify.
#
# juju-verify is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# juju-verify is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see https://www.gnu.org/licenses/.
"""Helper function to manage Juju action."""
from collections import OrderedDict
from typing import List, Any, Generator

from juju.action import Action


def data_from_action(action: Action, key: str, default: str = '') -> str:
    """Extract value from Action.data['results'] dictionary.

    :param action: juju.Action instance
    :param key: key to search for in action's results
    :param default: default value to return if the 'key' is not found
    :return: value from the action's results identified by 'key' or default
    """
    return action.data.get('results', {}).get(key, default)


class _Cache:
    """Cache class for action outputs."""

    def __init__(self, maxsize: int):
        """Initialize cache object."""
        self._cache: OrderedDict = OrderedDict()
        self.maxsize: int = maxsize

    def __getitem__(self, key: int) -> Any:
        """Get cached value."""
        if key in self._cache:
            self._cache.move_to_end(key)  # reorder cache

        return self._cache[key]

    def __setitem__(self, key: int, value: Any) -> None:
        """Cache the value using the key."""
        self._cache[key] = value

        # remove the oldest key
        if len(self._cache) > self.maxsize:
            oldest_key = next(iter(self._cache))
            del self._cache[oldest_key]

    def __iter__(self) -> Generator:
        """Iterate over cache keys."""
        for key in self._cache:
            yield key

    def clear(self) -> None:
        """Clear cached data."""
        self._cache.clear()

    @property
    def keys(self) -> List[Any]:
        """Return cached keys."""
        return list(self._cache.keys())


cache = _Cache(128)



# Copyright 2021 Canonical Limited.
#
# This file is part of juju-verify.
#
# juju-verify is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# juju-verify is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see https://www.gnu.org/licenses/.

"""Collection of juju verification exceptions."""


class VerificationError(Exception):
    """Exception related to the main verification process."""


class CharmException(Exception):
    """Exception related to the charm or the unit that runs it."""



# Copyright 2021 Canonical Limited.
#
# This file is part of juju-verify.
#
# juju-verify is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# juju-verify is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see https://www.gnu.org/licenses/.
"""Helper function to manage Juju unit."""
import asyncio
import os
import re
from typing import List, Dict, Any, Optional

from juju.action import Action
from juju.model import Model
from juju.unit import Unit

from juju_verify.exceptions import VerificationError, CharmException
from juju_verify.utils.action import cache

CHARM_URL_PATTERN = re.compile(r'^(.*):(.*/)?(?P<charm>.*)(-\d+)$')


def get_cache_key(unit: Unit, action: str, **params: Any) -> int:
    """Create hash key from unit, action and params."""
    return hash(
        hash(unit.entity_id) + hash(action) + hash(tuple(sorted(params.items())))
    )


async def run_action(unit: Unit, action: str, use_cache: bool = True,
                     params: Optional[Dict[str, Any]] = None) -> Action:
    """Run Juju action and wait for results."""
    params = params or {}
    key = get_cache_key(unit, action, **params)

    if key not in cache or not use_cache:
        _action = await unit.run_action(action, **params)
        result = await _action.wait()  # wait for result
        cache[key] = result  # save result to cache
        return result

    return cache[key]


def run_action_on_units(units: List[Unit], action: str, use_cache: bool = True,
                        params: Optional[Dict[str, Any]] = None) -> Dict[str, Action]:
    """Run Juju action on specified units.

    :param units: List/Tuple of Unit object
    :param action: Action to run on units
    :param use_cache: Use the cache to gather the result of the action
    :param params: Additional parameters for the action
    :return: Dict in format {unit_id: action} where unit_ids are strings
             provided in 'units' and actions are their matching,
             juju.Action objects that have been executed and awaited.
    """
    task_map = {unit.entity_id: run_action(unit, action, use_cache, params)
                for unit in units}
    loop = asyncio.get_event_loop()
    tasks = asyncio.gather(*task_map.values())
    results: List[Action] = loop.run_until_complete(tasks)

    result_map = dict(zip(task_map.keys(), results))

    failed_actions_msg = []
    for unit_id, action_result in result_map.items():
        if action_result.status != 'completed':
            failed_actions_msg.append(
                f'Action {action} (ID: {action_result.entity_id}) failed to complete '
                f'on unit {unit_id}. For more info see '
                f'"juju show-action-output {action_result.entity_id}"'
            )

    if failed_actions_msg:
        raise VerificationError(os.linesep.join(failed_actions_msg))

    return result_map


def run_action_on_unit(unit: Unit, action: str,  use_cache: bool = True,
                       params: Optional[Dict[str, Any]] = None) -> Action:
    """Run juju action on single unit.

    For more info, see docstring for 'run_action_on_units'. The only
    difference is that this function returns Action object directly, not
    dict {unit_id: action}.
    """
    results = run_action_on_units([unit], action, use_cache, params)
    return results[unit.entity_id]


def parse_charm_name(charm_url: str) -> str:
    """Parse charm name from full charm url.

    Example: 'cs:focal/nova-compute-141' -> 'nova-compute'
    """
    match = CHARM_URL_PATTERN.match(charm_url)
    if match is None:
        raise CharmException(f'Failed to parse charm-url: "{charm_url}"')
    return match.group('charm')


def verify_charm_unit(charm_name: str, *units: Unit) -> None:
    """Verify that units are based on required charm."""
    for unit in units:
        if not parse_charm_name(unit.charm_url) == charm_name:
            raise CharmException(f"The unit {unit.entity_id} does not belong to the "
                                 f"charm {charm_name}.")


def get_first_active_unit(units: List[Unit]) -> Optional[Unit]:
    """Find first unit in active workload status."""
    for unit in units:
        if unit.workload_status == "active":
            return unit

    return None


def get_applications_names(model: Model, application: str) -> List[str]:
    """Get all names of application based on the same charm."""
    applications = []
    for app_name, app in model.applications.items():
        unit = get_first_active_unit(app.units)
        if unit and parse_charm_name(unit.charm_url) == application:
            applications.append(app_name)

    return applications



# Copyright 2021 Canonical Limited.
#
# This file is part of juju-verify.
#
# juju-verify is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# juju-verify is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see https://www.gnu.org/licenses/.
"""Base for other modules that implement verification checks for specific charms."""
import asyncio
import logging
from collections import defaultdict
from collections import namedtuple
from typing import Callable, Dict, List, Optional, Any

from juju.action import Action
from juju.model import Model
from juju.unit import Unit

from juju_verify.exceptions import VerificationError
from juju_verify.utils.unit import run_action_on_units
from juju_verify.verifiers.result import aggregate_results, Result, Severity

logger = logging.getLogger(__name__)


class BaseVerifier:
    """Base class for implementation of verification checks for specific charms.

    Classes that inherit from this base must override class variable 'NAME' to
    match charm name (e.g. 'nova-compute') and override methods named
    `verify_<check_name>` with actual implementation of the  checks.

    NotImplemented exception will be raised if attempt is made to perform check
    that is not implemented in child class.
    """

    NAME = ''

    def __init__(self, units: List[Unit]):
        """Initiate verifier linked to the Juju units.

        All the checks that the verifier implements must expect that the action
        that is being verified is intended to be performed on all juju units
        in the 'self.units' simultaneously.

        :raises VerificationError: If 'units' parameter is empty
        :raises VerificationError: If 'units' parameter contains units from
                                   different models.
        """
        self.units = units
        self.affected_machines = set()
        models = set()

        if not self.units:
            raise VerificationError('Can not run verification. This verifier'
                                    ' is not associated with any units.')
        for unit in self.units:
            self.affected_machines.add(unit.machine.entity_id)
            models.add(unit.model)

        # Unit.model is mandatory property, so we end up either with one model
        # (correct) or multiple models (incorrect) in the 'models' set.
        if len(models) > 1:
            raise VerificationError('Verifier initiated with units from '
                                    'multiple models.')
        self.model: Model = models.pop()
        self._unit_ids: List[str] = []

    @property
    def unit_ids(self) -> List[str]:
        """Return entity IDs of self.units."""
        return self._unit_ids or [unit.entity_id for unit in self.units]

    @classmethod
    def supported_checks(cls) -> List[str]:
        """Return list of supported checks."""
        return list(cls._action_map().keys())

    @classmethod
    def _action_map(cls) -> Dict[str, Callable[['BaseVerifier'], Result]]:
        """Return verification checks mapper.

        The key is the verification name. The value is a callable method that
        implements the logic.
        """
        return {
            'shutdown': cls.verify_shutdown,
            'reboot': cls.verify_reboot,
        }

    def unit_from_id(self, unit_id: str) -> Unit:
        """Search self.units for unit that matches 'unit_id'.

        :param unit_id: ID of the unit to find
        :return: Unit that matches 'unit_id'
        """
        for unit in self.units:
            if unit.entity_id == unit_id:
                return unit
        raise VerificationError('Unit {} was not found in {} verifier.'
                                ''.format(unit_id, self.NAME))

    def check_affected_machines(self) -> Result:
        """Check if affected machines run other principal units.

        Log warning if machine that run units checked by this verifier also
        runs other principal units that are not being checked.
        """
        result = Result()
        machine_map: Dict = defaultdict(list)
        for unit in self.model.units.values():
            if unit.machine.entity_id in self.affected_machines \
                    and not unit.data.get('subordinate'):
                machine_map[unit.machine.entity_id].append(unit.entity_id)

        for machine, unit_list in machine_map.items():
            for unit in unit_list:
                if unit not in self.unit_ids:
                    result.add_partial_result(Severity.WARN,
                                              f'Machine {machine} runs other principal '
                                              f'unit that is not being checked: {unit}')
        return result

    def check_has_sub_machines(self) -> Result:
        """Check if the machine hosts containers or VMs.

        Logs warning if there are units running on sub machines that are children of the
        affected machines.
        """
        result = Result()
        ParentChildPair = namedtuple("ParentChildPair", "child parent")
        parent_child_pairs = {}
        parents = set()

        # Search for child machines
        for parent_unit in self.units:
            # search the list of units for any that have the unit"s machine as a parent
            for _, potential_child in self.model.units.items():
                if potential_child.machine.entity_id.startswith(
                        parent_unit.machine.entity_id + "/"):
                    parents.add(parent_unit.entity_id)
                    parent_child_pairs[potential_child.entity_id] = ParentChildPair(
                        child=potential_child,
                        parent=parent_unit
                    )

        task_map = {child_tag: parent_child_pair.child.is_leader_from_status()
                    for child_tag, parent_child_pair in parent_child_pairs.items()}

        loop = asyncio.get_event_loop()

        results = loop.run_until_complete(asyncio.gather(*task_map.values()))
        result_map = dict(zip(task_map.keys(), results))

        # loop through list of parents, format a message
        for check_unit_entity_id in parents:
            for _, child_parent_pair in parent_child_pairs.items():
                if child_parent_pair.parent.entity_id == check_unit_entity_id:
                    child_tag = child_parent_pair.child.entity_id
                    child_tag += "*" if result_map[child_tag] else ""
                    result.add_partial_result(Severity.WARN,
                                              f"{check_unit_entity_id} has units running"
                                              f" on child machines: {child_tag}")
        return result

    def verify(self, check: str) -> Result:
        """Execute requested verification check.

        :param check: Check to execute
        :return: Overall verification Result
        :raises NotImplementedError: If requested check is unsupported/unknown
        :raises VerificationError: If check fails in unexpected manner or if
                                   list of self.units is empty
        """
        preflight_results = aggregate_results(
            self.check_affected_machines(),
            self.check_has_sub_machines()
        )
        verify_action = self._action_map().get(check)
        if verify_action is None:
            raise NotImplementedError('Unsupported verification check "{}" for'
                                      ' charm {}'.format(check, self.NAME))

        try:
            logger.debug('Running check %s on units: %s', check, ','.join(self.unit_ids))
            main_results = verify_action(self)
            return aggregate_results(preflight_results, main_results)
        except NotImplementedError as exc:
            raise exc
        except Exception as exc:
            err = VerificationError('Verification failed: {}'.format(exc))
            raise err from exc

    def run_action_on_all(self, action: str, use_cache: bool = True,
                          params: Optional[Dict[str, Any]] = None) -> Dict[str, Action]:
        """Run juju action on all units in self.units.

        For more info, see docstring for 'run_action_on_units'.
        """
        return run_action_on_units(self.units, action, use_cache, params)

    def verify_shutdown(self) -> Result:
        """Child classes must override this method with custom implementation.

        'shutdown' check needs to be implemented on child classes.
        """
        raise NotImplementedError('Requested check "shutdown" is not '
                                  'implemented for "{}" '
                                  'charm.'.format(self.NAME))

    def verify_reboot(self) -> Result:
        """Child classes must override this method with custom implementation.

        'reboot' check needds to be implemented on child classes.
        """
        raise NotImplementedError('Requested check "reboot" is not '
                                  'implemented for "{}" '
                                  'charm.'.format(self.NAME))



# Copyright 2021 Canonical Limited.
#
# This file is part of juju-verify.
#
# juju-verify is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# juju-verify is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see https://www.gnu.org/licenses/.

"""Juju plugin to help verify the safety of maintenance operations.

The tool allows users to check whether it's safe to execute operations like
'shutdown' or 'reboot' on juju units without affecting availability and integrity.
"""
from .juju_verify import main



#!/usr/bin/env python3

# Copyright 2021 Canonical Limited.
#
# This file is part of juju-verify.
#
# juju-verify is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# juju-verify is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see https://www.gnu.org/licenses/.

"""Entrypoint to the 'juju-verify' plugin."""
import argparse
import logging
import os
import sys
import typing
from typing import List, Union

from juju import errors, loop
from juju.model import Model
from juju.unit import Unit

from juju_verify.exceptions import CharmException, VerificationError
from juju_verify.verifiers import BaseVerifier, get_verifier

logger = logging.getLogger(__package__)


def fail(err_msg: str) -> None:
    """Log error message and exit."""
    logger.error(err_msg)
    sys.exit(1)


async def find_units(model: Model, units: List[str]) -> List[Unit]:
    """Return list of juju.Unit objects that match with names in 'units' parameter.

    This function will exit program with error message if any of units is not
    found in the juju model.

    :param model: Juju model to search units in.
    :param units: List of unit names to search
    :return: List of matching juju.Unit objects
    """
    selected_units: List[Unit] = []

    for unit_name in units:
        unit = model.units.get(unit_name)
        if unit is None:
            fail('Unit "{}" not found in the model.'.format(unit_name))
        selected_units.append(unit)
    return selected_units


async def find_units_on_machine(model: Model, machines: List[str]) -> List[Unit]:
    """Find all principal units that run on selected machines.

    :param model: Juju model to search units in
    :param machines: names of Juju machines on which to search units
    :return: List of juju.Unit objects that match units running on the machines
    """
    return [unit for _, unit in model.units.items()
            if unit.machine.entity_id in machines
            and not unit.data.get("subordinate")]


async def connect_model(model_name: Union[str, None]) -> Model:
    """Connect to a custom or default Juju model.

    The Juju model can be identified by 'model_name' or the currently active
    model will be used if left unspecified.
    """
    model = Model()
    try:
        if model_name:
            logger.debug('Connecting to model "%s".', model_name)
            await model.connect_model(model_name)
        else:
            logger.debug('Connecting to currently active model.')
            await model.connect_current()
    except errors.JujuError as exc:
        fail("Failed to connect to the model.{}{}".format(os.linesep, exc))
    return model


class ExtendAction(argparse.Action):  # pylint: disable=too-few-public-methods
    """Extend action for argparse.

    NOTE (rgildein): This action should be removed after python 3.6 support ends, because
                     since Python 3.8 the "extend" is available directly in stdlib.
    """

    @typing.no_type_check
    def __call__(self, parser, namespace, values, option_string=None):
        """Extend existing items with values."""
        items = getattr(namespace, self.dest) or []
        items = [*items, *values]  # extend list of items
        setattr(namespace, self.dest, items)


def parse_args() -> argparse.Namespace:
    """Parse cli arguments."""
    description = "Verify that it's safe to perform selected action on " \
                  "specified units"
    parser = argparse.ArgumentParser(description=description)
    parser.register("action", "extend", ExtendAction)

    parser.add_argument('--model', '-m', required=False,
                        help='Connect to specific model.')
    parser.add_argument('check', choices=BaseVerifier.supported_checks(),
                        type=str.lower, help='Check to verify.')
    parser.add_argument('-l', '--log-level', type=str.lower,
                        help='Set amount of displayed information',
                        default='info', choices=['trace', 'debug', 'info'])

    target = parser.add_mutually_exclusive_group(required=True)
    target.add_argument('--units', '-u', action="extend", nargs='+', type=str,
                        help='Units to check.')
    target.add_argument('--machines', '-M', action="extend", nargs='+', type=str,
                        help='Check all units on the machine.')
    return parser.parse_args()


def config_logger(log_level: str) -> None:
    """Configure logging options."""
    log_level = log_level.lower()

    if log_level == 'trace':
        # 'trace' level enables debugging in juju lib and other dependencies
        logging.basicConfig(format='%(message)s', level=logging.DEBUG)
    else:
        # other levels apply only to juju-verify logging
        logging.basicConfig(format='%(message)s')
        if log_level == 'debug':
            logger.setLevel(logging.DEBUG)
        elif log_level == 'info':
            logger.setLevel(logging.INFO)
        else:
            fail('Unsupported log level requested: "{}"'.format(log_level))


def main() -> None:
    """Execute 'juju-verify' command."""
    args = parse_args()
    config_logger(args.log_level)
    model = loop.run(connect_model(args.model))
    units: List[Unit] = []

    if args.units:
        units = loop.run(find_units(model, args.units))
    elif args.machines:
        units = loop.run(find_units_on_machine(model, args.machines))
    else:
        fail('juju-verify must target either juju units or juju machines')

    try:
        verifier = get_verifier(units)
        result = verifier.verify(args.check)
        logger.info(str(result))
    except (CharmException, VerificationError, NotImplementedError) as exc:
        fail(str(exc))


if __name__ == '__main__':  # pragma: no cover
    main()



# Copyright 2021 Canonical Limited.
#
# This file is part of juju-verify.
#
# juju-verify is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# juju-verify is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see https://www.gnu.org/licenses/.

"""Package with classes implementing verification methods for various charms."""
import logging
import os
from typing import List

from juju.unit import Unit

from juju_verify.exceptions import CharmException
from juju_verify.utils.unit import parse_charm_name
from juju_verify.verifiers.base import BaseVerifier
from juju_verify.verifiers.ceph import CephOsd, CephMon
from juju_verify.verifiers.nova_compute import NovaCompute
from juju_verify.verifiers.result import Result, Severity

logger = logging.getLogger(__name__)


SUPPORTED_CHARMS = {
    'nova-compute': NovaCompute,
    'ceph-osd': CephOsd,
    'ceph-mon': CephMon,
}


def get_verifier(units: List[Unit]) -> BaseVerifier:
    """Implement Factory function "verifier" creator for the supplied units.

    :param units: Juju unit(s) for which you want to produce verifier
    :return: Correct verifier for given unit(s)
    :raises CharmException: Raised if units do not belong to the same charm or
        if the charm is unsupported for verification
    """
    charm_types = set()

    if not units:
        raise CharmException('List of units can not be empty when creating '
                             'verifier')
    for unit in units:
        charm_type = parse_charm_name(unit.data.get('charm-url', ''))
        logger.debug('Inferred charm for unit %s: %s', unit.entity_id,
                     charm_type)
        charm_types.add(charm_type)

    if len(charm_types) > 1:
        raise CharmException('Units are not running same charm. '
                             'Detected types: {}'.format(charm_types))

    charm = charm_types.pop()

    verifier = SUPPORTED_CHARMS.get(charm)

    if verifier is None:
        supported_charms = os.linesep.join(SUPPORTED_CHARMS.keys())
        raise CharmException('Charm "{}" is not supported by juju-verify. '
                             'Supported charms:{}'
                             '{}'.format(charm, os.linesep, supported_charms))
    logger.debug('Initiating verifier instance of class: %s',
                 verifier.__name__)
    return verifier(units=units)



# Copyright 2021 Canonical Limited.
#
# This file is part of juju-verify.
#
# juju-verify is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# juju-verify is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see https://www.gnu.org/licenses/.
"""ceph-osd verification."""
import json
import logging
from collections import defaultdict
from typing import Dict, Optional, Any, List

from juju.unit import Unit
from packaging.version import Version, InvalidVersion

from juju_verify.utils.action import data_from_action
from juju_verify.utils.unit import (
    verify_charm_unit,
    run_action_on_units,
    get_first_active_unit,
    get_applications_names
)
from juju_verify.verifiers.base import BaseVerifier
from juju_verify.verifiers.result import aggregate_results, Result, Severity
from juju_verify.exceptions import CharmException

logger = logging.getLogger()


class AvailabilityZone:
    """Availability zone."""

    def __init__(self, **data: str):
        """Availability zone initialization."""
        self._data = data

    def __getattr__(self, item: str) -> Optional[str]:
        """Get element from crush map hierarchy."""
        if item not in self.crush_map_hierarchy:
            raise AttributeError(f"'{self.__class__}' object has no attribute '{item}'")

        return self._data.get(item, None)

    def __eq__(self, other: object) -> bool:
        """Compare two Result instances."""
        if not isinstance(other, AvailabilityZone):
            return NotImplemented

        return str(self) == str(other)

    def __str__(self) -> str:
        """Return string representation of AZ objects."""
        return ",".join(
            [f"{crush_map_type}={self._data[crush_map_type]}" for crush_map_type
             in self.crush_map_hierarchy if crush_map_type in self._data]
        )

    def __hash__(self) -> int:
        """Return hash representation of AZ objects."""
        return hash(self.__str__())

    @property
    def crush_map_hierarchy(self) -> List[str]:
        """Get Ceph Crush Map hierarchy."""
        return [
            "root",  # 10
            "region",  # 9
            "datacenter",  # 8
            "room",  # 7
            "pod",  # 6
            "pdu",  # 5
            "row",  # 4
            "rack",  # 3
            "chassis",  # 2
            "host",  # 1
            "osd",  # 0
        ]


class CephCommon(BaseVerifier):  # pylint: disable=W0223
    """Parent class for CephMon and CephOsd verifier."""

    @classmethod
    def check_cluster_health(cls, *units: Unit) -> Result:
        """Check Ceph cluster health for specific units.

        This will execute `get-health` against each unit provided.

        :raises CharmException: if the units do not belong to the ceph-mon charm
        """
        verify_charm_unit("ceph-mon", *units)
        result = Result()
        action_map = run_action_on_units(list(units), "get-health")

        for unit, action in action_map.items():
            cluster_health = data_from_action(action, "message")
            logger.debug("Unit (%s): Ceph cluster health '%s'", unit, cluster_health)

            if "HEALTH_OK" in cluster_health and result.success:
                result.add_partial_result(Severity.OK,
                                          f"{unit}: Ceph cluster is healthy")
            elif "HEALTH_WARN" in cluster_health or "HEALTH_ERR" in cluster_health:
                result.add_partial_result(Severity.FAIL,
                                          f"{unit}: Ceph cluster is unhealthy")
            else:
                result.add_partial_result(Severity.FAIL,
                                          f"{unit}: Ceph cluster is in an unknown "
                                          f"state")

        if not action_map:
            result = Result(Severity.FAIL, "Ceph cluster is in an unknown state")

        return result

    @classmethod
    def get_replication_number(cls, unit: Unit) -> Optional[int]:
        """Get minimum replication number from ceph-mon unit.

        This function runs the `list-pools` action with the parameter 'detail=true'
        to get the replication number.
        :raises CharmException: if the unit does not belong to the ceph-mon charm
        :raises TypeError: if the object pools is not iterable
        :raises KeyError: if the pool detail does not contain `size` or `min_size`
        :raises json.decoder.JSONDecodeError: if json.loads failed
        """
        verify_charm_unit("ceph-mon", unit)
        action_map = run_action_on_units([unit], "list-pools", params={"format": "json"})
        action_output = data_from_action(action_map.get(unit.entity_id), "message", "[]")
        logger.debug("parse information about pools: %s", action_output)
        pools: List[Dict[str, Any]] = json.loads(action_output)

        if pools:
            return min(pool["size"] - pool["min_size"] for pool in pools)

        return None

    @classmethod
    def get_number_of_free_units(cls, unit: Unit) -> int:
        """Get number of free units from ceph df.

        The `ceph df` will provide a cluster’s data usage and data distribution among
        pools and this function calculates the number of units that are safe to remove.
        """
        verify_charm_unit("ceph-mon", unit)
        # NOTE (rgildein): This functionality is not complete and will be part of the
        #                  fix on LP#1921121.
        logger.warning("WARNING: The function to get the number of free units from "
                       "'ceph df' is in WIP and returns only 1. See LP#1921121 "
                       "for more information.")
        return 1

    @classmethod
    def get_availability_zones(cls,  *units: Unit) -> Dict[str, AvailabilityZone]:
        """Get information about availability zones for ceph-osd units."""
        # NOTE (rgildein): This has been tested, but it will be fully functional only
        #                  after merging the changes.
        #                  https://review.opendev.org/c/openstack/charm-ceph-osd/+/778159
        verify_charm_unit("ceph-osd", *units)
        action_map = run_action_on_units(list(units), "get-availability-zone")

        availability_zone = {}
        for unit, action in action_map.items():
            action_output = data_from_action(action, "availability-zone", "{}")
            unit_availability_zone = json.loads(action_output)["unit"]
            unit_availability_zone.pop("host")  # need to compare AZ without host
            availability_zone[unit] = AvailabilityZone(**unit_availability_zone)

        return availability_zone


class CephOsd(CephCommon):
    """Implementation of verification checks for the ceph-osd charm."""

    NAME = "ceph-osd"

    def __init__(self, units: List[Unit]):
        """Ceph-osd charm verifier."""
        super().__init__(units=units)
        self._ceph_mon_app_map: Optional[Dict[str, Unit]] = None

    @property
    def ceph_mon_app_map(self) -> Dict[str, Unit]:
        """Get a map between ceph-osd applications and the first ceph-mon unit.

        :returns: Dictionary with keys as distinct applications of verified units and
                  values as the first ceph-mon unit obtained from the relation with the
                  ceph-mon application (<application_name>:mon).
        """
        if self._ceph_mon_app_map is None:
            self._ceph_mon_app_map = self._get_ceph_mon_app_map()

        return self._ceph_mon_app_map

    def _get_ceph_mon_unit(self, app_name: str) -> Optional[Unit]:
        """Get first ceph-mon unit from relation."""
        try:
            for relation in self.model.applications[app_name].relations:
                if relation.matches(f"{app_name}:mon"):
                    return get_first_active_unit(relation.provides.application.units)

        except (IndexError, KeyError) as error:
            logger.debug("Error to get ceph-mon unit from relations: %s", error)

        return None

    def _get_ceph_mon_app_map(self) -> Dict[str, Unit]:
        """Get first ceph-mon units related to verified units.

        This function groups by distinct application names for verified units, and then
        finds the relation ("<application>:mon") between the application and ceph-mon.
        The first unit of ceph-mon will be obtained from this relation.
        :returns: Map between verified and ceph-mon units
        """
        applications = {unit.application for unit in self.units}
        logger.debug("affected applications %s", map(str, applications))

        app_map = {name: self._get_ceph_mon_unit(name) for name in applications}
        logger.debug("found units %s", map(str, app_map.values()))

        return {name: unit for name, unit in app_map.items() if unit is not None}

    def get_free_app_units(self, apps: List[str]) -> Dict[str, int]:
        """Get number of free units for each application."""
        free_units = {}
        for app_name in apps:
            ceph_mon_unit = self._get_ceph_mon_unit(app_name)
            if ceph_mon_unit:
                free_units[app_name] = self.get_number_of_free_units(ceph_mon_unit)

        return free_units

    def get_apps_availability_zones(
            self, apps: List[str]) -> Dict[AvailabilityZone, List[Unit]]:
        """Get information about availability zone for each ceph-osd unit in application.

        This function return dictionary contain AZ as key and list of units as value.
        e.g. {AZ(per_host=False, rack="nova", row=None): [<Unit entity_id="ceph-osd/0">]}
        """
        availability_zones = defaultdict(list)
        for app_name in apps:
            app_units = self.model.applications[app_name].units
            app_availability_zones = self.get_availability_zones(*app_units)
            for unit_id, availability_zone in app_availability_zones.items():
                availability_zones[availability_zone].append(self.model.units[unit_id])

        return availability_zones

    def check_ceph_cluster_health(self) -> Result:
        """Check Ceph cluster health for unique ceph-mon units from ceph_mon_app_map."""
        unique_ceph_mon_units = set(self.ceph_mon_app_map.values())
        return self.check_cluster_health(*unique_ceph_mon_units)

    def check_replication_number(self) -> Result:
        """Check the minimum number of replications for related applications."""
        result = Result()

        for app_name, ceph_mon_unit in self.ceph_mon_app_map.items():
            min_replication_number = self.get_replication_number(ceph_mon_unit)
            if min_replication_number is None:
                continue  # get_replication_number returns None if no pools are available

            units = {
                unit.entity_id for unit in self.units if unit.application == app_name
            }
            inactive_units = {
                unit.entity_id for unit in self.model.applications[app_name].units
                if unit.workload_status != "active"
            }

            if len(units.union(inactive_units)) > min_replication_number:
                result.add_partial_result(
                    Severity.FAIL,
                    f"The minimum number of replicas in '{app_name}' is "
                    f"{min_replication_number:d} and it's not safe to restart/shutdown "
                    f"{len(units):d} units. {len(inactive_units):d} units are not "
                    f"active."
                )

        if result.empty:
            result.add_partial_result(Severity.OK,
                                      'Minimum replica number check passed.')
        return result

    def check_availability_zone(self) -> Result:
        """Check availability zones resources.

        This function checks whether the units can be shutdown/reboot without
        interrupting operation in the availability zone.
        """
        result = Result()
        ceph_osd_apps = get_applications_names(self.model, "ceph-osd")
        free_app_units = self.get_free_app_units(ceph_osd_apps)
        availability_zones = self.get_apps_availability_zones(ceph_osd_apps)

        for availability_zone, units in availability_zones.items():
            inactive_units = {
                unit.entity_id for unit in units if unit.workload_status != "active"
            }
            units_to_remove = {
                unit.entity_id for unit in units if unit.entity_id in self.unit_ids
            }
            free_units = min(free_app_units[unit.application] for unit in units)
            logger.debug("The availability zone %s has %d inactive and %d free units. "
                         "Trying to remove %s units.", str(availability_zone),
                         len(inactive_units), free_units, units_to_remove)

            if len(units_to_remove.union(inactive_units)) > free_units:
                result += Result(
                    Severity.FAIL,
                    f"It's not safe to removed units {units_to_remove} in the "
                    f"availability zone '{availability_zone}'. [free_units="
                    f"{free_units:d}, inactive_units={len(inactive_units):d}]"
                )

        if result.empty:
            result.add_partial_result(Severity.OK,
                                      'Availability zone check passed.')
        return result

    def verify_reboot(self) -> Result:
        """Verify that it's safe to reboot selected ceph-osd units."""
        return aggregate_results(self.check_ceph_cluster_health(),
                                 self.check_replication_number(),
                                 self.check_availability_zone())

    def verify_shutdown(self) -> Result:
        """Verify that it's safe to shutdown selected ceph-osd units."""
        return self.verify_reboot()


class CephMon(CephCommon):
    """Implementation of verification checks for the ceph-mon charm."""

    NAME = "ceph-mon"

    def verify_reboot(self) -> Result:
        """Verify that it's safe to reboot selected ceph-mon units."""
        version_check = self.check_version()
        if not version_check.success:
            return version_check

        # Get one ceph-mon unit per each application
        app_map = {unit.application: unit for unit in self.units}
        unique_app_units = app_map.values()

        return aggregate_results(
            version_check,
            self.check_quorum(),
            self.check_cluster_health(*unique_app_units)
        )

    def verify_shutdown(self) -> Result:
        """Verify that it's safe to shutdown selected units."""
        return self.verify_reboot()

    def check_quorum(self) -> Result:
        """Check that the shutdown does not result in <50% mons alive."""
        result = Result()

        action_name = "get-quorum-status"
        action_results = self.run_action_on_all(action_name)

        affected_hosts = {unit.machine.hostname for unit in self.units}

        for unit_id, action in action_results.items():
            # run this per unit because we might have multiple clusters
            known_mons = set(json.loads(data_from_action(action, "known-mons")))
            online_mons = set(json.loads(data_from_action(action, "online-mons")))

            mon_count = len(known_mons)
            mons_after_change = len(online_mons - affected_hosts)

            if mons_after_change <= mon_count // 2:
                result.add_partial_result(Severity.FAIL, f"Removing unit {unit_id} will"
                                                         f" lose Ceph mon quorum")

        if result.empty:
            result.add_partial_result(Severity.OK, 'Ceph-mon quorum check passed.')

        return result

    def check_version(self) -> Result:
        """Check minimum required version of juju agents on units.

        Ceph-mon verifier requires that all the units run juju agent >=2.8.10 due to
        reliance on juju.Machine.hostname feature.
        """
        min_version = Version('2.8.10')
        result = Result()
        for unit in self.units:
            juju_version = unit.safe_data.get('agent-status', {}).get('version', '')
            try:
                if Version(juju_version) < min_version:
                    fail_msg = (f'Juju agent on unit {unit.entity_id} has lower than '
                                f'minumum required version. {juju_version} < '
                                f'{min_version}')
                    result.add_partial_result(Severity.FAIL, fail_msg)
            except InvalidVersion as exc:
                raise CharmException(f'Failed to parse juju version from '
                                     f'unit {unit.entity_id}.') from exc

        if result.empty:
            result.add_partial_result(Severity.OK, 'Minimum juju version check passed.')

        return result



# Copyright 2021 Canonical Limited.
#
# This file is part of juju-verify.
#
# juju-verify is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# juju-verify is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see https://www.gnu.org/licenses/.
"""nova-compute verification."""
import json
import logging

from juju_verify.utils.action import data_from_action
from juju_verify.utils.unit import run_action_on_unit
from juju_verify.verifiers.base import BaseVerifier
from juju_verify.verifiers.result import aggregate_results, Result, Severity

logger = logging.getLogger()


class NovaCompute(BaseVerifier):
    """Implementation of verification checks for the nova-compute charm."""

    NAME = 'nova-compute'

    def check_no_running_vms(self) -> Result:
        """Check that none of the units have VMs running on them."""
        result = Result()
        instance_count_action = 'instance-count'
        instance_count_results = self.run_action_on_all(instance_count_action)

        for unit_id, action in instance_count_results.items():
            running_vms = int(data_from_action(action, 'instance-count'))
            if running_vms != 0:
                result.add_partial_result(Severity.FAIL, f'Unit {unit_id} is running '
                                                         f'{running_vms} VMs.')
            else:
                result.add_partial_result(Severity.OK, f'Unit {unit_id} is running '
                                                       f'{running_vms} VMs.')
        return result

    def check_no_empty_az(self) -> Result:
        """Check that removing units wont cause empty availability zone."""
        def is_active(node: dict) -> bool:
            return node['state'] == 'up' and node['status'] == 'enabled'

        node_name_actions = self.run_action_on_all('node-name')
        target_nodes = [data_from_action(action, 'node-name')
                        for _, action in node_name_actions.items()]

        action = run_action_on_unit(self.units[0], 'list-compute-nodes')
        compute_nodes = json.loads(data_from_action(action, 'compute-nodes'))

        affected_zones = {node['zone'] for node in compute_nodes
                          if node['host'] in target_nodes}
        zones_after_change = {node['zone'] for node in compute_nodes
                              if node['host'] not in target_nodes and
                              is_active(node)}

        empty_zones = affected_zones - zones_after_change

        if empty_zones:
            result = Result(Severity.FAIL, f'Removing these units would leave following'
                                           f' availability zones empty: {empty_zones}')
        else:
            result = Result(Severity.OK, 'Empty Availability Zone check passed.')
        return result

    def verify_reboot(self) -> Result:
        """Verify that it's safe to reboot selected nova-compute units."""
        return aggregate_results(self.check_no_running_vms(),
                                 self.check_no_empty_az())

    def verify_shutdown(self) -> Result:
        """Verify that it's safe to shutdown selected nova-compute units."""
        return self.verify_reboot()
